"""From the file dumps from Stud.IP generate the database of all students."""

import argparse
import sqlite3
import contextlib

import pandas as pd
from hashing import numeric_hash, SALT, HASH_LENGTH
from studip import BLACKLIST

# These are the columns that will go into the database.
KEEP_COLUMNS = [
    'Anrede',
    'Vorname',
    'Nachname',
    'Nutzernamen',
    'E-Mail',
    'Anmeldedatum',
    'Hashes',
]


def get_commandline_args():
    """Read arguments from command line."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--files',
        '-f',
        nargs='*',  # could be more than one file
        dest='student_files',
        required=True,
        action='store',
        help='Participants files from Stud.IP.',
    )
    parser.add_argument(
        '--dbfile',
        '-db',
        dest='database_file',
        required=True,
        action='store',
        help='Path to the database file.',
    )
    return parser.parse_args()


def merge_files_and_cleanup(input_files):
    """Merge the files, remove duplicates and remove lecturers."""
    # In case multiple files have been passed we must first merge these and then remove duplicate entries. It
    # appears to be more stable to use the student ID to find duplicates rather than to take the whole entry
    # into account.
    df = pd.concat(
        [pd.read_csv(f, delimiter=';') for f in input_files]
    ).drop_duplicates(['Nutzernamen'])

    # Remove the lecturers from the list of students.
    index_list = [
        (surname == df['Nachname']).loc[lambda m: m].index.values[0]
        for surname in BLACKLIST
    ]
    df.drop(index_list, inplace=True)

    # Remove some column names.
    df.drop(columns=list(set(df.columns) - set(KEEP_COLUMNS)), inplace=True)

    return df


def add_hashes(df):
    df['Hashes'] = [
        numeric_hash(SALT, HASH_LENGTH, sid) for sid in df['Nutzernamen']
    ]
    return df


class Database:
    def __init__(self, data, conn):
        self.data = data
        self.conn = conn

    def dump(self, table_name):
        """Write to table in database."""
        self.data.to_sql(
            table_name, self.conn, if_exists='replace', index=False
        )

    def fetch_student(self, name, table_name):
        """Fetch data for specific student from some table in database."""
        db_query_string = f'SELECT * FROM {table_name}'
        db_query_string += ' ' + f"WHERE Nachname='{name}'"
        # https://stackoverflow.com/questions/9885217/in-python-if-i-return-inside-a-with-block-will-the-file-still-close
        return pd.read_sql_query(db_query_string, self.conn)


def main():
    cmd_args = get_commandline_args()

    # Since the may be more than just one file dumped from Stud.IP we must merge these into one
    # dataframe with duplicates removed.
    df = merge_files_and_cleanup(cmd_args.student_files)

    # We now need to add the hashes based on the student IDs ("Nutzername" column in tha table).
    df = add_hashes(df)

    # Dump student data to database.
    with contextlib.closing(sqlite3.connect(cmd_args.database_file)) as conn:
        db = Database(df, conn)
        db.dump(table_name='student_data')

    with contextlib.closing(sqlite3.connect(cmd_args.database_file)) as conn:
        db = Database(df, conn)
        data = db.fetch_student(name='Al Aji', table_name='student_data')
        print(data)


if __name__ == '__main__':
    main()
