"""nbgrader exporter plugin to get grades in a nice tabular form"""
import itertools
import json
import logging
import pathlib

import nbgrader.api
import numpy as np
import pandas as pd
from nbgrader.plugins import ExportPlugin

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def find_grade(grade_book, cell, notebook, assignment, student):
    try:
        return grade_book.find_grade(
            cell.name, notebook.name, assignment.name, student.id
        ).score
    except nbgrader.api.MissingEntry:
        return pd.NA


def has_entry(gradebook, assignment, missing_ok, student=None, notebook=None):
    if missing_ok:
        return True
    if student:
        if notebook:
            return notebook.name in [
                notebook.name
                for notebook in gradebook.find_submission(
                    assignment.name, student.id
                ).notebooks
            ]
        else:
            return assignment.name in [
                assignment.name
                for assignment in gradebook.student_submissions(student.id)
            ]
    else:
        return gradebook.assignment_submissions(assignment.name)


def export_to_multiindex(gradebook, students, assignments, missing_ok):
    """Export the grades as a 2D table for printing or uploading"""

    grades_dict = {
        student.id: {
            (assignment.name, notebook.name, cell_index): find_grade(
                gradebook, cell, notebook, assignment, student
            )
            for assignment in assignments
            if has_entry(gradebook, assignment, missing_ok, student)
            for notebook in assignment.notebooks
            if has_entry(gradebook, assignment, student, notebook)
            for cell_index, cell in enumerate(
                itertools.chain(notebook.grade_cells, notebook.task_cells)
            )
        }
        for student in students
    }

    grades_dict['reachable'] = {
        (assignment.name, notebook.name, cell_index): cell.max_score
        for assignment in assignments
        for notebook in assignment.notebooks
        for cell_index, cell in enumerate(
            itertools.chain(notebook.grade_cells, notebook.task_cells)
        )
    }

    grades_table = pd.DataFrame.from_dict(grades_dict).fillna(np.NaN)
    grades_table.index = pd.MultiIndex.from_tuples(
        grades_table.index, names=('Assignment', 'Task', 'Cell')
    )
    return grades_table


def export_to_df(gradebook, students, assignments, missing_ok):
    """Export the grades as a 2D table for printing or uploading"""
    # TODO: Deduplicate: let csv exporters use the multiindex as well

    logger.info(f'Exporting students: {", ".join(students) or "all"}')
    students = [
        gradebook.find_student(student) for student in students
    ] or gradebook.students

    logger.info(
        f'Exporting assignments: {", ".join(map(str, assignments)) or "all"}'
    )
    assignments = [
        gradebook.find_assignment(assignment) for assignment in assignments
    ] or gradebook.assignments

    grades_dict = {
        student.id: {
            (assignment.name, notebook.name, cell_index): find_grade(
                gradebook, cell, notebook, assignment, student
            )
            for assignment in assignments
            if has_entry(gradebook, assignment, missing_ok, student)
            for notebook in assignment.notebooks
            if has_entry(gradebook, assignment, student, notebook)
            for cell_index, cell in enumerate(
                itertools.chain(notebook.grade_cells, notebook.task_cells)
            )
        }
        for student in students
    }

    grades_dict['reachable'] = {
        (assignment.name, notebook.name, cell_index): cell.max_score
        for assignment in assignments
        for notebook in assignment.notebooks
        for cell_index, cell in enumerate(notebook.grade_cells)
    }

    grades_table = pd.DataFrame.from_dict(grades_dict, orient='index')

    grades_table.columns = [
        f'{assignment.name}.'
        if len(assignments) > 1
        else '' f'{nb_index + 1}.{cell_index + 1}'
        for assignment in assignments
        if has_entry(gradebook, assignment, missing_ok)
        for nb_index, notebook in enumerate(assignment.notebooks)
        for cell_index, cell in enumerate(notebook.grade_cells)
    ]

    return grades_table


def export_assignments_to_csv(
    gradebook, students, assignments, filename, missing_ok
):
    """Export the aggregate points as a 2D table for printing or uploading"""
    # TODO: Deduplicate and use the multiindex as well

    students = [
        gradebook.find_student(student) for student in students
    ] or gradebook.students
    assignments = [
        gradebook.find_assignment(assignment) for assignment in assignments
    ] or gradebook.assignments

    grades_dict = {
        student.id: {
            assignment.name: sum(
                find_grade(gradebook, cell, notebook, assignment, student)
                for notebook in assignment.notebooks
                if has_entry(gradebook, assignment, student, notebook)
                for cell in notebook.grade_cells
            )
            for assignment in assignments
            if has_entry(gradebook, assignment, missing_ok, student)
        }
        for student in students
    }

    grades_dict['reachable'] = {
        assignment.name: sum(
            cell.max_score
            for notebook in assignment.notebooks
            for cell in notebook.grade_cells
        )
        for assignment in assignments
    }

    grades_table = pd.DataFrame.from_dict(grades_dict, orient='index')

    grades_table.columns = [
        assignment.name
        for assignment in assignments
        if has_entry(gradebook, assignment, missing_ok)
    ]

    grades_table = grades_table.dropna(how='all', axis=0)

    logger.info(f'exporting to {filename}')
    grades_table.to_csv(filename.resolve(), header=True, index_label='Kennung')


class DetailedCompleteExporter(ExportPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.to = self.to or 'grades.csv'

    def export(self, gradebook):
        grades_table = export_to_df(
            gradebook,
            self.student,
            self.assignment,
            missing_ok=True,
        )
        logger.info(f'exporting to {self.to}')
        grades_table.to_csv(
            self.to.resolve(), header=True, index_label='Kennung'
        )


class DetailedSparseExporter(ExportPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.to = self.to or 'grades.csv'

    def export(self, gradebook):
        grades_table = export_to_df(
            gradebook,
            self.student,
            self.assignment,
            missing_ok=False,
        )

        logger.info(f'exporting to {self.to}')
        grades_table.to_csv(
            self.to.resolve(), header=True, index_label='Kennung'
        )


class SMTPConnection:
    def __init__(self, sender):
        import getpass
        import ssl

        self.smtp_server = sender.get('server') or input(
            'Type your SMTP server and press enter: '
        )

        self.port = int(
            sender.get('port')
            or input(
                f'On which port does the SMTP server {self.smtp_server} answer (default: 587)? '
            )
            or 587
        )

        self.sender_login = sender.get('username') or input(
            f'Type your username for {self.smtp_server} and press enter: '
        )

        self.sender_email = sender.get('email') or input(
            f'Type the e-mail address associated with {self.sender_login}@{self.smtp_server} and press enter: '
        )

        self.password = getpass.getpass(
            'Type your password for '
            f'{self.sender_login}@{self.smtp_server} '
            'and press enter: '
        )

        self.context = ssl.create_default_context()

    def __enter__(self):
        import smtplib

        self.conn = smtplib.SMTP(self.smtp_server, self.port)
        self.conn.ehlo()
        self.conn.starttls(context=self.context)
        self.conn.ehlo()
        self.conn.login(self.sender_login, self.password)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.quit()


def compose_mime_message(sender, recipient, subject, body, attachments):
    from email import encoders
    from email.mime.base import MIMEBase
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText

    message = MIMEMultipart()
    message['From'] = sender
    message['To'] = recipient
    message['Subject'] = subject
    message['Bcc'] = sender

    message.attach(MIMEText(body, 'plain'))

    for path in attachments:
        with open(path, 'rb') as file:
            part = MIMEBase('application', 'octet-stream')
            part.set_payload(file.read())
        encoders.encode_base64(part)

        part.add_header(
            'Content-Disposition',
            f'attachment; filename= {path.name}',
        )

        message.attach(part)

    return message.as_string()


class MailExporter(ExportPlugin):
    """Send feedback by mail.

    Requires additional configuration files:
    ADDRESS_BOOK -- json with name - mail address pairs
    MAIL_TEMPLATE -- template for assignment feedback
    """

    TEMPLATE_PATH = pathlib.Path('MailData')
    ADDRESS_BOOK: pathlib.Path
    MAIL_TEMPLATE: pathlib.Path
    missing_ok: bool
    attach_feedback: bool
    points_100_percent: int
    klausurzulassung: dict

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        with open(self.ADDRESS_BOOK, 'r') as f:
            self.address_book = json.load(f)

        with open(self.MAIL_TEMPLATE, 'r') as f:
            self.subject = (
                f.readline().removeprefix('subject: ').removesuffix('\n')
            )
            self.message = f.read()
        self.log.info(
            'Exporting feedback with subject: \n'
            f'"{self.subject}"\n'
            'and message template: """\n'
            f'{self.message}\n"""'
        )

        sender = self.to.split(',', 1)[0]
        self.sender = self.address_book.get(sender)
        if self.sender is None:
            self.log.error(
                f'Sender "{sender}" not recognized.'
                f' Please add to "{self.ADDRESS_BOOK}"'
                ' and pass as first argument to `self.to`.'
            )

        prompt = f'Sending feedback as {self.sender["name"]} to {self.to}. Is this correct [y/n]? '
        valid = {'yes': True, 'y': True, 'no': False, 'n': False}
        while (response := input(prompt)) not in valid:
            print('Please respond with "yes" or "no"')
        if not valid[response]:
            """Please select your name in the `self.to` flag. Exiting."""
            exit(1)

    def export(self, gradebook):
        students = (
            list(map(gradebook.find_student, self.student))
            or gradebook.students
        )
        assignments = (
            list(map(gradebook.find_assignment, self.assignment))
            or gradebook.assignments
        )

        grades_table = export_to_multiindex(
            gradebook,
            students,
            assignments,
            missing_ok=self.missing_ok,
        )

        course_root = pathlib.Path(self.config['CourseDirectory']['root'])
        feedback_dir = self.config['CourseDirectory']['directory_structure']
        feedback_files = f'{course_root}/{feedback_dir}/{{notebook_name}}.html'
        reachable = grades_table['reachable']
        if reachable.apply(float.is_integer).all():
            reachable = reachable.astype(int)

        format_assignment = {
            'assignment_id': '", "'.join(a.name for a in assignments),
            'reachable': reachable.sum(),
            'sender': self.sender,
        }

        with SMTPConnection(self.sender['smtp']) as smtp:
            for student_id, results in grades_table.iteritems():
                if student_id == 'reachable':
                    continue
                student = gradebook.find_student(student_id)

                if results.fillna(0.0).apply(float.is_integer).all():
                    results = results.astype(
                        'Int64'
                    )  # use nullable integer value

                details = (
                    results.astype(str) + '/' + reachable.astype(str)
                ).unstack(fill_value='')
                details.columns.name = ''
                details.columns = 'Teil ' + (details.columns + 1).astype(str)
                details.index.names = ['Übung', 'Aufgabe']

                address_book = {
                    key: value['smtp']['email']
                    for key, value in self.address_book.items()
                } | {'students': gradebook.find_student(student_id).email}

                if self.attach_feedback:
                    feedback_paths = [
                        pathlib.Path(
                            feedback_files.format(
                                nbgrader_step='feedback',
                                student_id=student_id,
                                assignment_id=assignment.name,
                                notebook_name=notebook.name,
                            )
                        ).resolve()
                        for assignment in assignments
                        for notebook in assignment.notebooks
                    ]

                    attachments = list(
                        filter(pathlib.Path.is_file, feedback_paths)
                    )
                    if not self.missing_ok and not attachments:
                        self.log.info(
                            f'Skipping student {student.id} ({student.first_name} {student.last_name}):'
                            f'No feedback files found in {feedback_paths}.'
                        )
                        continue
                else:
                    attachments = []

                format_student = {
                    'first_name': student.first_name,
                    'last_name': student.last_name,
                    'reached': results.sum(),
                    'details': details,
                }

                if self.points_100_percent:
                    template = self.klausurzulassung[
                        2 * results.sum() > self.points_100_percent
                    ]
                    with open(template, mode='r') as f:
                        format_student['klausurzulassung'] = f.read().format(
                            points_100_percent=self.points_100_percent
                        )

                body = self.message.format(
                    **format_assignment,
                    **format_student,
                )

                message = compose_mime_message(
                    sender=smtp.sender_email,
                    recipient=address_book['students'],
                    subject=self.subject,
                    body=body,
                    attachments=attachments,
                )

                to_addrs = list(
                    f'{{{self.to}}}'.replace(',', '};{')
                    .format_map(address_book)
                    .split(';')
                )

                if not to_addrs:
                    self.log.info(
                        f'Recipient not {"recognized" if self.to else "given"}.'
                    )
                    self.log.info(
                        f'Dry run: not sending to {address_book["students"]}.'
                    )
                    continue

                self.log.info(f'Sending e-mail to {to_addrs}.')
                smtp.conn.sendmail(
                    from_addr=smtp.sender_email,
                    to_addrs=to_addrs,
                    msg=message,
                )


class MailTaskExporter(MailExporter):
    TEMPLATE_PATH = MailExporter.TEMPLATE_PATH
    ADDRESS_BOOK = TEMPLATE_PATH / 'address_book.json'
    MAIL_TEMPLATE = TEMPLATE_PATH / 'task_feedback.txt'
    missing_ok = False
    attach_feedback = True
    points_100_percent = 0


class MailSummaryExporter(MailExporter):
    TEMPLATE_PATH = MailExporter.TEMPLATE_PATH
    ADDRESS_BOOK = TEMPLATE_PATH / 'address_book.json'
    MAIL_TEMPLATE = TEMPLATE_PATH / 'semester_summary.txt'
    missing_ok = True
    attach_feedback = False
    points_100_percent = 120
    klausurzulassung = {
        False: TEMPLATE_PATH / 'Klausurzulassung' / 'final_missed.txt',
        True: TEMPLATE_PATH / 'Klausurzulassung' / 'final_reached.txt',
    }


class AggregateExporter(ExportPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.to = self.to or 'grades.csv'

    def export(self, gradebook):
        return export_assignments_to_csv(
            gradebook,
            self.student,
            self.assignment,
            pathlib.Path(self.to),
            missing_ok=True,
        )
