import csv
import json
import os
import shutil

import dateutil.parser as DP
import nbgrader.plugins.zipcollect
from nbgrader.utils import unzip


class StudIPExtractor(nbgrader.plugins.zipcollect.ExtractorPlugin):

    info_file = 'archive_filelist.csv'
    students_dict_file = (
        'ClassroomData/SoSe_2022/StudentData/nbgrader_ids.json'
    )

    def extract(self, archive_path, extracted_path):
        """Extracts top-level zip files in-place and renames the contained directories."""

        archives = [
            zipfile
            for zipfile in os.listdir(archive_path)
            if os.path.splitext(zipfile)[1] == '.zip'
        ]

        studip_zip_base = archives.pop()
        if archives:
            self.log.error(
                f'More than one file found in directory: {archive_path}\n'
                'Please provide exactly one archive from stud.IP.'
            )
            return

        studip_zip_full = os.path.join(archive_path, studip_zip_base)

        studip_extracted_dir, ext = os.path.splitext(studip_zip_full)
        if ext not in self.zip_ext:
            self.log.warning(f'Non-zipfile found: {studip_zip_base}')
            return

        unzip(
            studip_zip_full,
            archive_path,
            zip_ext=self.zip_ext,
            create_own_folder=True,
            tree=False,
        )

        self._rename(studip_extracted_dir)
        self._extract(studip_extracted_dir, extracted_path)
        shutil.rmtree(studip_extracted_dir)

    def _rename(self, studip_files):
        """Rename submissions from archive_filelist.csv and pick out the newest one."""

        if self.info_file not in os.listdir(studip_files):
            self.log.warning(
                'Cannot find archive filelist. Not renaming files.'
            )
            return

        with open(self.students_dict_file) as f:
            students_dict = json.load(f)

        with open(os.path.join(studip_files, self.info_file)) as f:
            reader = csv.DictReader(f, delimiter=';')
            for line in reader:
                name = f'{line["Nachname"]}, {line["Vorname"]}'
                self.log.info(f'Found file from student {name}')
                if name not in students_dict:
                    self.log.warn(f'No id found for student {name}')
                    continue
                submission_date = DP.parse(line['Datum'], dayfirst=True)

                moved_path = os.path.join(
                    studip_files,
                    f'StudIP_student_{students_dict[name]["id"]}_{submission_date.isoformat()}.zip',
                )
                student_file = os.path.join(studip_files, line['Pfad'])
                if os.path.exists(moved_path):
                    self.log.warn(
                        f'Found duplicate submission from student {name}'
                    )
                    if students_dict[name]['date'] > submission_date:
                        shutil.rmtree(student_file)
                        continue
                    else:
                        students_dict[name]['date'] = submission_date
                else:
                    students_dict[name]['date'] = submission_date

                shutil.move(student_file, os.path.join(moved_path))

    def _extract(self, archive_path, extracted_path):
        if not os.listdir(archive_path):
            self.log.warning(f'No files found in directory: {archive_path}')
            return

        for root, _, archive_files in os.walk(archive_path):
            if not archive_files:
                continue

            extract_to = os.path.normpath(
                os.path.join(
                    extracted_path, os.path.relpath(root, archive_path)
                )
            )
            if not os.path.isdir(extract_to):
                os.makedirs(extract_to)

            for zfile in archive_files:
                zfile = os.path.join(root, zfile)
                filename, ext = os.path.splitext(os.path.basename(zfile))
                # unzip (tree) each archive file in archive_path
                if ext in self.zip_ext:
                    # double splitext for .tar.gz
                    fname, ext = os.path.splitext(os.path.basename(filename))
                    if ext == '.tar':
                        filename = fname
                    self.log.info(
                        f'Extracting {filename} to {os.path.relpath(root, extract_to)}'
                    )
                    unzip(
                        zfile,
                        extract_to,
                        zip_ext=self.zip_ext,
                        create_own_folder=True,
                        tree=True,
                    )

                # move each non-archive file in archive_path
                else:
                    dest = os.path.join(extract_to, os.path.basename(zfile))
                    self.log.info(
                        f'Copying {os.path.basename(zfile)} to {dest}'
                    )
                    shutil.copy(zfile, dest)


#    def find_newest_submissions(self, prefix_length: int, submissions_dir: pathlib.Path = pathlib.Path()):
#       logger.info(f'Discovering submissions in {submissions_dir.resolve()}')

#       newest_submissions = {}
#       for file in submissions_dir.iterdir():
#           if not is_zipfile(file):
#               logger.warning(
#                   f'Non-zipfile submission was ignored: {file.resolve()}'
#               )
#               continue

#           if not (sub_id := str(file.name)[:prefix_length]).isnumeric():
#               logger.warning(
#                   f'Submission with non-numeric ID was ignored: {file.resolve()}'
#               )
#               continue

#           if sub_id in newest_submissions:
#               logger.warning(f'Non-unique submission with ID {sub_id} found!')
#               modification_time = file.stat().st_mtime
#               if modification_time < newest_submissions[sub_id].stat().st_mtime:
#                   continue

#           newest_submissions[sub_id] = file

#       logger.info(
#           f'Total {len(newest_submissions)} unique submissions discovered.'
#       )
#       return newest_submissions
